@@ -1,8 +1,8 @@
-import typing
-
 import pandas
 import iso8601
 import io
+import datetime
+from typing import Callable, Iterable, Any, Tuple, Dict, Union, Optional, TextIO, List, cast, Generator
 import json
 
 from . import server, utils
@@ -16,7 +16,7 @@ class API(server.Server):
           and speaks a compatible API version, then you can use instance methods to interact with the API server.
     '''
 
-    def __init__(self, *args, fetch_batch_size=1000, **kwargs):
+    def __init__(self, *args, fetch_batch_size:int=1000, **kwargs):
         '''Initializes an instance of the API class that let's you communicate with an EdelweissData server.
 
         :param base_url: Base url of the EdelweissData API server to communicate with
@@ -28,7 +28,7 @@ class API(server.Server):
         self.get('/health')
         self.get('/ready')
 
-    def _fetch_in_batches(self, fetch, limit, offset):
+    def _fetch_in_batches(self, fetch : Callable[[Optional[int], Optional[int]], Any], limit: Optional[int], offset: Optional[int]) -> Iterable[Any]:
         start = 0 if offset is None else offset
         end = None if limit is None else limit + start
         batch_size = self.fetch_batch_size
@@ -59,21 +59,21 @@ class API(server.Server):
             else:
                 return
 
-    def openapi_documents(self):
+    def openapi_documents(self) -> List[str]:
         '''Returns a list of all dataset specific openapi descriptions (i.e. one openapi document for each dataset with the
              precise Json Schema of the particular datasets data endpoint).
 
         :returns: A list of url strings at which to retrieve the openapi.json documents for the documents
         '''
-        return self.get('/openapidocuments')
+        return cast(List[str], self.get('/openapidocuments'))
 
-    def openapi(self):
+    def openapi(self) -> Dict[str, Any]:
         '''Returns the OpenAPI definition of the entire EdelweissData REST API.
 
         :returns: The OpenAPI definition as a dict'''
-        return self.get('/openapi.json')
+        return cast(Dict[str, Any] , self.get('/openapi.json'))
 
-    def get_in_progress_datasets(self, limit=None, offset=None):
+    def get_in_progress_datasets(self, limit: Optional[int]=None, offset: Optional[int]=None) -> List["InProgressDataset"]:
         '''Returns a list of all in-progress datasets you are allowed to access (needs authentication).
 
         :returns: a list of InProgressDatasets
@@ -82,7 +82,7 @@ class API(server.Server):
         :param offset: Starting offset from which to retrieve the datasets
         '''
         route = '/datasets/in-progress'
-        def fetch(limit, offset):
+        def fetch(limit: Optional[int], offset: Optional[int]):
             request = {'limit': limit, 'offset': offset}
             return self.get(route, json=request)
         datasets = []
@@ -90,7 +90,7 @@ class API(server.Server):
             datasets += [InProgressDataset.decode(d, api=self) for d in batch]
         return datasets
 
-    def get_in_progress_dataset(self, id):
+    def get_in_progress_dataset(self, id: str) -> "InProgressDataset":
         '''Returns an in-progress datasets with a given id.
 
         :returns: an InProgressDataset
@@ -99,7 +99,7 @@ class API(server.Server):
         route = '/datasets/{}/in-progress'.format(id)
         return InProgressDataset.decode(self.get(route), api=self)
 
-    def create_in_progress_dataset(self, name):
+    def create_in_progress_dataset(self, name: str) -> "InProgressDataset":
         '''Creates a new in-progress dataset on the server and returns it.
 
         :returns: The InProgressDataset that was created.
@@ -108,7 +108,7 @@ class API(server.Server):
         route = '/datasets/create'
         return InProgressDataset.decode(self.post(route, {'name': name}), api=self)
 
-    def get_raw_datasets(self, columns=None, condition=None, include_description=None, include_schema=None, include_metadata=None, include_aggregations=None, aggregation_filters=None, limit=None, offset=None, order_by=None, ascending=True, latest_only=None):
+    def get_raw_datasets(self, columns: Optional[Iterable[Tuple[str,str]]]=None, condition: Optional["QueryExpression"] =None, include_description: Optional[bool]=None, include_schema: Optional[bool]=None, include_metadata: Optional[bool]=None, include_aggregations: Optional[bool]=None, aggregation_filters:Optional[Dict[str, Iterable[str]]]=None, limit:Optional[int]=None, offset:Optional[int]=None, order_by:Optional[List["QueryExpression"]]=None, ascending:Union[bool, List[bool]]=True, latest_only: Optional[bool]=None) -> Any:
         '''Get the published datasets. Unlike the more high-level get_published_datasets this method
              does not create a dataframe but returns the raw list of dicts representing the json response.
              Unless explicity included the fields schema, metadata and description will not be included
@@ -147,7 +147,7 @@ class API(server.Server):
           version of each dataset
         '''
         route = '/datasets'
-        request = {}
+        request: Dict[str, Any] = {}
         if columns is not None:
             request['columns'] = [
                 {
@@ -178,7 +178,7 @@ class API(server.Server):
             request['latestOnly'] = latest_only
         return self.get(route, json=request)
 
-    def get_published_datasets(self, columns=None, condition=None, include_description=False, include_schema=False, include_metadata=False, aggregation_filters=None, limit=None, offset=None, order_by=None, ascending=True, dataset_column_name='dataset', latest_only=None):
+    def get_published_datasets(self, columns: Optional[Iterable[Tuple[str,str]]]=None, condition: Optional["QueryExpression"] =None, include_description: Optional[bool]=None, include_schema: Optional[bool]=None, include_metadata: Optional[bool]=None, include_aggregations: Optional[bool]=None, aggregation_filters:Optional[Dict[str, Iterable[str]]]=None, limit:Optional[int]=None, offset:Optional[int]=None, order_by:Optional[List["QueryExpression"]]=None, ascending:Union[bool, List[bool]]=True, dataset_column_name: str='dataset', latest_only: Optional[bool]=None) -> pandas.DataFrame:
         '''Returns a dataframe of all published datasets that match query.
 
         :returns: a dataframe indexed by the id and version, which in addition
@@ -254,7 +254,7 @@ class API(server.Server):
         else:
             return pandas.DataFrame(columns=column_names, index=index)
 
-    def get_published_dataset_aggregations(self, columns=None, condition=None, aggregation_filters=None):
+    def get_published_dataset_aggregations(self, columns:Optional[Iterable[Tuple[str,str]]]=None, condition: Optional["QueryExpression"]=None, aggregation_filters:Optional[Dict[str, Iterable[str]]]=None) -> pandas.Series:
         '''Returns aggregation buckets and their sizes for each column.
 
         :returns: aggregations as a Series with an index of buckets and terms, for example
@@ -276,7 +276,7 @@ class API(server.Server):
         )
         return utils.decode_aggregations(response['aggregations'])
 
-    def get_published_dataset(self, id, version=None):
+    def get_published_dataset(self, id: str, version:Optional[Union[str,int]]=None) -> "PublishedDataset":
         '''Returns a published dataset with a given id and version.
 
         :returns: the PublishedDataset
@@ -289,10 +289,10 @@ class API(server.Server):
         route = '/datasets/{}/versions/{}'.format(id, version)
         return PublishedDataset.decode(self.get(route), api=self)
 
-    def get_published_dataset_versions(self, id):
+    def get_published_dataset_versions(self, id: str) -> List["PublishedDataset"]:
         '''Returns all published versions of dataset with a given id.
 
-        :returns: a list of dicts containing id, version and name for each version of the dataset
+        :returns: a list of PublishedDatasets
         :param id: id of the dataset'''
         route = '/datasets/{}'.format(id)
         response = self.get(route)
@@ -302,7 +302,7 @@ class API(server.Server):
             for version in versions
         ]
 
-    def create_in_progress_dataset_from_csv_file(self, name: str, file: typing.TextIO, metadata: dict = None):
+    def create_in_progress_dataset_from_csv_file(self, name: str, file: TextIO, metadata: dict = None) -> "InProgressDataset":
         '''Creates a new in-progress dataset from a CSV file on the server.
 
         :returns: the updated dataset
@@ -316,7 +316,7 @@ class API(server.Server):
             dataset.upload_metadata(metadata)
         return dataset
 
-    def create_published_dataset_from_csv_file(self, *args, changelog='Initial version', **kwargs):
+    def create_published_dataset_from_csv_file(self, *args, changelog: str='Initial version', **kwargs) -> "PublishedDataset":
         '''Creates a new published dataset from a CSV file on the server.
 
         :returns: the published dataset
@@ -328,7 +328,7 @@ class API(server.Server):
         published_dataset = dataset.publish(changelog)
         return published_dataset
 
-    def get_dataset_permissions(self, dataset_id):
+    def get_dataset_permissions(self, dataset_id: str) -> "DatasetPermissions":
         '''Get the permissions for the given dataset id
 
         :returns: the DatasetPermissions instance for this dataset
@@ -337,7 +337,7 @@ class API(server.Server):
         route = '/datasets/{}/permissions'.format(dataset_id)
         return DatasetPermissions.decode(self.get(route))
 
-    def add_dataset_user_permission(self, dataset_id, user):
+    def add_dataset_user_permission(self, dataset_id: str, user: "DatasetPermissions.User"):
         '''Add a user to a dataset
 
         :param dataset_id: the id of the dataset
@@ -346,7 +346,7 @@ class API(server.Server):
         route = '/datasets/{}/permissions/users/add'.format(dataset_id)
         return self.post(route, user.encode())
 
-    def remove_dataset_user_permission(self, dataset_id, email):
+    def remove_dataset_user_permission(self, dataset_id: str, email: str):
         '''Remove a user from a dataset
 
         :param dataset_id: the id of the dataset
@@ -358,7 +358,7 @@ class API(server.Server):
         }
         return self.post(route, payload)
 
-    def add_dataset_group_permission(self, dataset_id, group):
+    def add_dataset_group_permission(self, dataset_id: str, group: "DatasetPermissions.Group"):
         '''Add a group to a dataset
 
         :param dataset_id: the id of the dataset
@@ -367,7 +367,7 @@ class API(server.Server):
         route = '/datasets/{}/permissions/groups/add'.format(dataset_id)
         return self.post(route, group.encode())
 
-    def remove_dataset_group_permission(self, dataset_id, name):
+    def remove_dataset_group_permission(self, dataset_id: str, name: str):
         '''Remove a group from a dataset
 
         :param dataset_id: the id of the dataset
@@ -379,7 +379,7 @@ class API(server.Server):
         }
         return self.post(route, payload)
 
-    def change_dataset_visibility(self, dataset_id, is_public):
+    def change_dataset_visibility(self, dataset_id: str, is_public: bool):
         '''Set if the dataset should be public or access protected when published
 
         :param dataset_id: the id of the dataset
@@ -396,6 +396,7 @@ class API(server.Server):
         return self.get('/oidc')
 
 
+
 class Schema:
     '''The schema of the dataset describing the columns (name, description, datatype, rdf predicate, ...)
 
@@ -403,7 +404,7 @@ class Schema:
     class Column:
         '''The schema data of one column. This tells EdelweissData the name of the column, the datatype to use, how to handle missing values, ...
         '''
-        def __init__(self, name, description, data_type, array_value_separator, missing_value_identifiers, indices, rdf_predicate, statistics, visible):
+        def __init__(self, name: str, description: str, data_type: str, array_value_separator: str, missing_value_identifiers: Iterable[str], indices: List[str], rdf_predicate: str, statistics: Any, visible: bool):
             self.name = name
             self.description = description
             self.data_type = data_type
@@ -444,7 +445,7 @@ class Schema:
                 'visible': self.visible,
             }
 
-    def __init__(self, columns):
+    def __init__(self, columns: Iterable[Column]):
         self.columns = columns
 
     def __repr__(self):
@@ -469,7 +470,7 @@ class DatasetPermissions:
          an is_public field that indicates whether unauthenticated users can see this dataset when published.
     '''
     class User:
-        def __init__(self, email, can_write):
+        def __init__(self, email: str, can_write: bool):
             self.email = email
             self.can_write = can_write
 
@@ -484,7 +485,7 @@ class DatasetPermissions:
             }
 
     class Group:
-        def __init__(self, name, can_write):
+        def __init__(self, name: str, can_write: bool):
             self.name = name
             self.can_write = can_write
 
@@ -498,7 +499,7 @@ class DatasetPermissions:
                 'canWrite': self.can_write
             }
 
-    def __init__(self, id, users, groups, is_public):
+    def __init__(self, id: str, users: List[User], groups: List[Group], is_public: bool):
         self.id = id
         self.users = users
         self.groups = groups
@@ -525,10 +526,15 @@ class DatasetPermissions:
         }
 
 
+
+
+
+
+
 class InProgressDataset:
     '''InProgressDataset - datasets that are not yet published and for which data can be uploaded, the schema modified, metadata changed etc.
     '''
-    def __init__(self, id, name, schema, created, description, metadata, data_source, api):
+    def __init__(self, id: str, name: str, schema: Optional[Schema], created: datetime.datetime, description: str, metadata: dict, data_source: Optional["PublishedDataset"], api: "API"):
         self.id = id
         self.name = name
         self.schema = schema
@@ -565,19 +571,19 @@ class InProgressDataset:
             'datasource': self.metadata,
         }
 
-    def sample(self):
+    def sample(self) -> List[List[str]]:
         '''Retrieve a list of lists representing a sample of the tabular data of this dataset. This
             includes only a sample (e.g. the first N rows) of the data so that they can be displayed to a
             user as an example or similar.
         '''
         route = '/datasets/{}/in-progress/sample'.format(self.id)
-        return self.api.get(route)
+        return cast(List[List[str]], self.api.get(route))
 
-    def update(self, name: typing.Optional[str] = None,
-                     description: typing.Optional[str] = None,
-                     data_source: typing.Optional['PublishedDataset'] = None,
-                     schema: typing.Optional[Schema] = None,
-                     metadata: typing.Optional[dict] = None):
+    def update(self, name: Optional[str] = None,
+                     description: Optional[str] = None,
+                     data_source: Optional['PublishedDataset'] = None,
+                     schema: Optional[Schema] = None,
+                     metadata: Optional[dict] = None):
         '''Update various attributes of a in-progress dataset. All parameters are options; those that are
         None will not have their values changed.
 
@@ -609,7 +615,7 @@ class InProgressDataset:
         '''
         return self.update(schema=schema)
 
-    def upload_schema_file(self, file: typing.TextIO):
+    def upload_schema_file(self, file: TextIO):
         '''Upload a schema file (an open text file containing the schema in Json form).
 
         :param file: The open text file to upload the schema from
@@ -617,14 +623,14 @@ class InProgressDataset:
         schema = Schema.decode(json.load(file))
         return self.update(schema=schema)
 
-    def upload_metadata(self, metadata):
+    def upload_metadata(self, metadata: dict):
         '''Upload metadata (as a dict, not a file).
 
         :param schema: The metadata to upload
         '''
         return self.update(metadata=metadata)
 
-    def upload_metadata_file(self, file: typing.TextIO):
+    def upload_metadata_file(self, file: TextIO):
         '''Upload a metadata file (an open text file containing the metadata in Json form).
 
         :param file: The open text file to upload the metadata from
@@ -632,7 +638,7 @@ class InProgressDataset:
         metadata = json.load(file)
         return self.update(metadata=metadata)
 
-    def upload_data(self, data):
+    def upload_data(self, data: TextIO):
         '''Upload tabular data (a CSV file)
 
         :param data: An open text file containing the csv data to upload
@@ -640,28 +646,29 @@ class InProgressDataset:
         route = '/datasets/{}/in-progress/data/upload'.format(self.id)
         return self.api.upload(route, {'data': data})
 
-    def upload_dataframe_data(self, dataframe):
+    def upload_dataframe_data(self, dataframe: pandas.DataFrame):
         '''Upload a pandas dataframe as the data content into an InProgress dataset
 
         :param dataframe: A Pandas dataframe containing the data to upload
         '''
         data = io.StringIO()
         dataframe.to_csv(data, index=False)
-        return self.upload_data(data.getvalue())
+        data.seek(0)
+        return self.upload_data(data)
 
     def set_description(self, description: str):
         '''Set the description of the dataset. The description is assumed to be markdown formatted text, similar to a Github README.md
         '''
         return self.update(description=description)
 
-    def set_name(self, name):
+    def set_name(self, name: str):
         '''Set the name of the dataset.
         '''
         route = '/datasets/{}/in-progress'.format(self.id)
         self.api.post(route, json={'name': name})
         self.name = name
 
-    def set_data_source(self, dataset):
+    def set_data_source(self, dataset: "PublishedDataset"):
         '''Set the data source for an in-progress dataset. This allows you to efficiently re-use the data of a PublishedDataset
             to create a new dataset without re-uploading the data. It is also useful if you want to create a new version of a
             PublishedDataset to fix a mistake in the metadata or description.
@@ -669,7 +676,7 @@ class InProgressDataset:
         :param dataset: the PublishedDataset to copy data from when publishing
         '''
         route = '/datasets/{}/in-progress'.format(self.id)
-        data_source = {'id': dataset.id, 'version': dataset.version}
+        data_source = PublishedDataset.decode({'id': dataset.id, 'version': dataset.version}, self.api)
         return self.update(data_source=data_source)
 
     def infer_schema(self):
@@ -685,7 +692,7 @@ class InProgressDataset:
         route = '/datasets/{}/in-progress'.format(self.id)
         return self.api.delete(route)
 
-    def publish(self, changelog):
+    def publish(self, changelog: str) -> "PublishedDataset":
         '''Attempts to publish the dataset. This means that a new version of a PublishedDataset will be created (and returned by this call)
             and this InProgressDataset is no longer useable.
 
@@ -693,7 +700,7 @@ class InProgressDataset:
         route = '/datasets/{}/in-progress/publish'.format(self.id)
         return PublishedDataset.decode(self.api.post(route, {'changelog': changelog}), api=self.api)
 
-    def copy_from(self, published_dataset):
+    def copy_from(self, published_dataset: "PublishedDataset"):
         '''Copies all content from a PublishedDataset to this InProgressDataset. Useful to create new versions. See also set_data_source for
             a more lightweight operation if you don't need to change the data or schema structure.
         '''
@@ -713,7 +720,7 @@ class PublishedDataset:
     '''
     LATEST = 'LATEST'
 
-    def __init__(self, id, version, name, schema, created, description, metadata, row_count, is_public, api):
+    def __init__(self, id: str, version: int, name: str, schema: Optional[Schema], created: datetime.datetime, description: Optional[str], metadata: dict, row_count: Optional[int], is_public: bool, api: "API"):
         self.id = id
         self.version = version
         self.name = name
@@ -736,7 +743,7 @@ class PublishedDataset:
         self._is_public = dataset.is_public
 
     @property
-    def schema(self):
+    def schema(self) -> Optional[Schema]:
         '''Schema of this dataset. If this PublishedDatset instance was loaded from an API.get_published_datasets query and
              include_schema was not set to true then this property will be lazy loaded from the server when it is first accessed.
         '''
@@ -745,7 +752,7 @@ class PublishedDataset:
         return self._schema
 
     @property
-    def description(self):
+    def description(self) -> Optional[str]:
         '''Description text of this dataset. If this PublishedDatset instance was loaded from an API.get_published_datasets query and
              include_description was not set to true then this property will be lazy loaded from the server when it is first accessed.
         '''
@@ -754,7 +761,7 @@ class PublishedDataset:
         return self._description
 
     @property
-    def metadata(self):
+    def metadata(self) -> Any:
         '''Metadata of this dataset. If this PublishedDatset instance was loaded from an API.get_published_datasets query and
              include_metadata was not set to true then this property will be lazy loaded from the server when it is first accessed.
         '''
@@ -801,7 +808,7 @@ class PublishedDataset:
             'isPublic': self.is_public,
         }
 
-    def new_version(self):
+    def new_version(self) -> "InProgressDataset":
         '''Create a new version of this PublishedDataset. This will create and return a new InProgressDataset
             that can be filled with content by uploading new files or copying data from a PublishedDataset
 
@@ -810,7 +817,7 @@ class PublishedDataset:
         route = '/datasets/{}/versions/{}/create-new-version'.format(self.id, self.version)
         return InProgressDataset.decode(self.api.post(route), api=self.api)
 
-    def get_raw_data(self, columns=None, condition=None, include_aggregations=None, aggregation_filters=None, limit=None, offset=None, order_by=None, ascending=True):
+    def get_raw_data(self, columns: Optional[Iterable[str]]=None, condition: Optional["QueryExpression"]=None, include_aggregations: Optional[bool]=None, aggregation_filters:Optional[Dict[str, Iterable[str]]]=None, limit: Optional[int]=None, offset: Optional[int]=None, order_by: Optional[List["QueryExpression"]]=None, ascending:Union[bool, List[bool]]=True):
         '''Gets the raw tabular data JSON response for a PublishedDataset. The data can be filtered so that only required columns or rows
             are retrieved.
 
@@ -835,7 +842,7 @@ class PublishedDataset:
           it must be of the same length as the order_by list, and the order is
           the ascending/descending for each particular component.'''
         route = '/datasets/{}/versions/{}/data'.format(self.id, self.version)
-        request = {}
+        request: Dict[str, Any] = {}
         if columns is not None:
             request['columns'] = columns
         if condition is not None:
@@ -852,7 +859,7 @@ class PublishedDataset:
             request['orderBy'] = utils.encode_order_by(order_by, ascending)
         return self.api.get(route, json=request)
 
-    def get_data(self, columns=None, condition=None, aggregation_filters=None, limit=None, offset=None, order_by=None, ascending=True):
+    def get_data(self, columns: Optional[Iterable[str]]=None, condition: Optional["QueryExpression"]=None, aggregation_filters:Optional[Dict[str, Iterable[str]]]=None, limit: Optional[int]=None, offset: Optional[int]=None, order_by: Optional[List["QueryExpression"]]=None, ascending:Union[bool, List[bool]]=True) -> pandas.DataFrame:
         '''Gets the (tabular) data of a PublishedDataset as a pandas Dataframe. The data can be filtered so that only required columns or rows
             are retrieved.
 
@@ -899,7 +906,7 @@ class PublishedDataset:
         else:
             return pandas.DataFrame(columns=column_names, index=ids)
 
-    def get_data_aggregations(self, columns=None, condition=None, aggregation_filters=None):
+    def get_data_aggregations(self, columns:Optional[Iterable[str]]=None, condition:Optional["QueryExpression"]=None, aggregation_filters:Optional[Dict[str, Iterable[str]]]=None) -> pandas.Series:
         '''Returns aggregation buckets and their sizes for each column.
 
         :returns: aggregations as a Series with an index of buckets and terms, for example
@@ -927,7 +934,7 @@ class PublishedDataset:
         )
         return utils.decode_aggregations(response['aggregations'])
 
-    def openapi(self):
+    def openapi(self) -> Any:
         '''Returns a OpenAPI descriptions for the data endpoint of this PublishedDataset, taking the schema
             and thus the precise JSON structure of the response into account.
 
@@ -945,7 +952,8 @@ class PublishedDataset:
         '''Deletes all versions of a published dataset
         '''
         route = '/datasets/{}'.format(self.id)
-        return self.api.delete(route)
+        self.api.delete(route)
+
 
 
 class QueryExpression:
@@ -982,14 +990,14 @@ class QueryExpression:
             return '{0}({1!r})'.format(type(self).__name__, self.value)
 
     @classmethod
-    def _convert_if_necessary(cls, value):
+    def _convert_if_necessary(cls, value) -> "QueryExpression":
         if isinstance(value, cls):
             return value
         else:
             return cls(value)
 
     @classmethod
-    def decode(cls, value):
+    def decode(cls, value) -> "QueryExpression":
         if isinstance(value, dict) and len(value) == 1:
             name, arguments = value.popitem()
             return cls(name, [cls.decode(argument) for argument in arguments])
@@ -1003,7 +1011,7 @@ class QueryExpression:
             return self.value
 
     @classmethod
-    def search_anywhere(cls, term):
+    def search_anywhere(cls, term: str) -> "QueryExpression":
         '''Constructs a SearchAnywhere expression. Only rows will be returned that contain the search term in one of their text-like columns.
 
         :param term: The string to search for in all text-like columns.
@@ -1011,7 +1019,7 @@ class QueryExpression:
         return cls('searchAnywhere', [cls(term)])
 
     @classmethod
-    def column(cls, column_name):
+    def column(cls, column_name: str) -> "QueryExpression":
         '''Constructs a Column expression.
 
         :param column_name: the name of the column
@@ -1019,7 +1027,7 @@ class QueryExpression:
         return cls('column', [cls(column_name)])
 
     @classmethod
-    def system_column(cls, column_name):
+    def system_column(cls, column_name: str) -> "QueryExpression":
         '''Constructs a SystemColumn expression. SystemColumns are special columns maintained by EdelweissData.
             The following SystemColumns are available:
                 name (text): the name of a dataset
@@ -1031,7 +1039,7 @@ class QueryExpression:
         return cls('systemColumn', [cls(column_name)])
 
     @classmethod
-    def exact_search(cls, expr, term):
+    def exact_search(cls, expr: "QueryExpression", term: str) -> "QueryExpression":
         '''Constructs an ExactSearch expression. Only rows where the expr expression exactly matches the term will be returned. This can be used
             to match exact substrings or exact numerical values
 
@@ -1040,7 +1048,7 @@ class QueryExpression:
         return cls('exactSearch', [expr, cls(term)])
 
     @classmethod
-    def fuzzy_search(cls, expr, term):
+    def fuzzy_search(cls, expr: "QueryExpression", term: str) -> "QueryExpression":
         '''Constructs a FuzzySearch expression. Only rows where the expr expression fuzzy-matches the term will be returned. Fuzzy-matching
             uses trigram indexing to match slightly different spellings.
 
@@ -1050,7 +1058,7 @@ class QueryExpression:
         return cls('fuzzySearch', [expr, cls(term)])
 
     @classmethod
-    def substructure_search(cls, substructure, superstructure):
+    def substructure_search(cls, substructure: "QueryExpression", superstructure: "QueryExpression") -> "QueryExpression":
         '''Constructs a SubstructureSearch expression that uses chemical substructure testing. Only rows where the chemical substructure is contained in
             the chemical superstructure are returned.
 
@@ -1060,7 +1068,7 @@ class QueryExpression:
         return cls('substructureSearch', [cls._convert_if_necessary(substructure), cls._convert_if_necessary(superstructure)])
 
     @classmethod
-    def tanimoto_similarity(cls, left, right):
+    def tanimoto_similarity(cls, left: "QueryExpression", right: "QueryExpression") -> "QueryExpression":
         '''Calculates the tanimoto distance between two molecular fingerprints.
 
         :param left: the left argument. Often a SMILES string constant value or Column of datatype SMILES.
@@ -1069,7 +1077,7 @@ class QueryExpression:
         return cls('tanimotoSimilarity', [cls._convert_if_necessary(left), cls._convert_if_necessary(right)])
 
     @classmethod
-    def cast(cls, expr, data_type):
+    def cast(cls, expr: "QueryExpression", data_type: str) -> "QueryExpression":
         '''Creates a Cast expression. This attempts to convert one datatype into another.
 
         :param expr: The expression to cast
@@ -1078,7 +1086,7 @@ class QueryExpression:
         return cls('cast', [expr, cls(data_type)])
 
     @classmethod
-    def contains(cls, expr, element):
+    def contains(cls, expr: "QueryExpression", element: "QueryExpression") -> "QueryExpression":
         '''Creates a Contains expression. Tests if an expression contains an element. Often used
             to check if columns of an Array datatype contain a value.
 
@@ -1088,7 +1096,7 @@ class QueryExpression:
         return cls('contains', [expr, cls(element)])
 
     @classmethod
-    def contained_in(cls, expr, element):
+    def contained_in(cls, expr: "QueryExpression", element: "QueryExpression") -> "QueryExpression":
         '''Creates a ContainedIn expression. Tests if an expression is contained in an element. Often used
             to check if columns of an Array datatype are contained in a value.
 
@@ -1100,32 +1108,32 @@ class QueryExpression:
     def __getitem__(self, json_path):
         return self.json_path_query(self, json_path)
 
-    def __and__(self, other):
+    def __and__(self, other: "QueryExpression") -> "QueryExpression":
         return type(self)('and', [self, self._convert_if_necessary(other)])
 
-    def __rand__(self, other):
+    def __rand__(self, other: "QueryExpression") -> "QueryExpression":
         return type(self)('and', [self._convert_if_necessary(other), self])
 
-    def __or__(self, other):
+    def __or__(self, other: "QueryExpression") -> "QueryExpression":
         return type(self)('or', [self, self._convert_if_necessary(other)])
 
-    def __ror__(self, other):
+    def __ror__(self, other: "QueryExpression") -> "QueryExpression":
         return type(self)('and', [self._convert_if_necessary(other), self])
 
-    def __invert__(self):
+    def __invert__(self: "QueryExpression") -> "QueryExpression":
         return type(self)('not', [self])
 
     def __eq__(self, other):
         return type(self)('eq', [self, self._convert_if_necessary(other)])
 
-    def __lt__(self, other):
+    def __lt__(self, other) -> "QueryExpression":
         return type(self)('lt', [self, self._convert_if_necessary(other)])
 
-    def __le__(self, other):
+    def __le__(self, other) -> "QueryExpression":
         return type(self)('le', [self, self._convert_if_necessary(other)])
 
-    def __gt__(self, other):
+    def __gt__(self, other) -> "QueryExpression":
         return type(self)('gt', [self, self._convert_if_necessary(other)])
 
-    def __ge__(self, other):
+    def __ge__(self, other) -> "QueryExpression":
         return type(self)('ge', [self, self._convert_if_necessary(other)])
